#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#define N_Gr 900		// 10000
#define N_Go 100		// 900
#define N_MF 100		// 500
#define T 500.0			// duration of the CS
#define T_Offset 50.0		// starts -50 ms prior to the CS onset
#define DT 0.02
#define N_Step 50		// (1/DT)
#define N_Gr_MF 3		// # of connections Gr <- MF
#define N_Gr_Go 3		// # of connections Gr <- Go
#define N_Go_MF 20		// # of connections Go <- MF
#define N_Go_Gr 100		// # of connections Go <- Gr

// Parameters (Table 1, p44)
const double Tau_Gr_MF = 2.86;
const double Tau_Gr_Go = 5.0;
const double Tau_Go_MF = 2.87;
const double Tau_Go_Gr = 2.86;
const double Tau_Gr_Thr = 1.7;
const double Tau_Go_Thr = 2.0;
const double E_leak = -60.0;
const double E_ex = 0;
const double E_inh = -80.0;
const double g_Gr_leak = 0.07;
const double g_Go_leak = 0.07;
const double W_Gr_MF = 40.0;	// 0.15;
const double W_Gr_Go = 0.15;
const double W_Go_MF = 0;	// 0.007;
const double W_Go_Gr = 0.5;	// 0.008;
const double BaseThr_Gr = -40.0;
const double BaseThr_Go = -35.0;
const double MaxThr_Gr = -35.0;
const double MaxThr_Go = -25.0;

//
int **id_Go_MF, **id_Go_Gr, **id_Gr_MF, **id_Gr_Go, **spike;
double dg_Go_MF[N_Go], g_Go_MF[N_Go], dg_Go_Gr[N_Go], g_Go_Gr[N_Go];
double dg_Gr_MF[N_Gr], g_Gr_MF[N_Gr], dg_Gr_Go[N_Gr], g_Gr_Go[N_Gr];
double dV_Go[N_Go], V_Go[N_Go], dV_Gr[N_Gr], V_Gr[N_Gr];
double Thr_Go[N_Go], Thr_Gr[N_Gr];
int S_Go[N_Go], S_Gr[N_Gr], S_MF[N_MF];

extern void init_genrand(unsigned long);
extern double genrand_real2(void);

// Eq.(2.3), p43
void compute_dg(int N, double *dg, int nconn, int **id, int *S,
		double W, double *g, double tau)
{
  int i, j, k;

  for(i = 0; i < N; i++){
    dg[i] = 0;
    for(k = 0; k < nconn; k++){
      j = id[i][k];
      dg[i] += DT*(S[j] * W * (1 - g[i]) - g[i]/tau);
    }
  }
}

void update_g(int N, double *dg, double *g)
{
  int i;

  for(i = 0; i < N; i++){
    g[i] += dg[i];
  }
}

// Eq.(2.1), p43
void compute_dV(int N, double *dV, double *V, double g_leak,
		double *g1, double E1, double *g2, double E2)
{
  int i = 0;

  for(i = 0; i < N; i++){
    dV[i] = DT*(-g_leak * (V[i] - E_leak)
		-g1[i] * (V[i] - E1)
		-g2[i] * (V[i] - E2));
  }
}

void update_V(int N, double *dV, double *V)
{
  int i;

  for(i = 0; i < N; i++){
    V[i] += dV[i];
  }
}

// Eq.(2.2), p43
void generate_spike(int N, double *V, int *S, double *Thr,
		    double BaseThr, double MaxThr, double Tau_Thr)
{
  int i;
  const double decay = exp(-DT/Tau_Thr);

  for(i = 0 ; i < N; i++){
    if (V[i] >= Thr[i]){
      S[i] = 1;
      Thr[i] = MaxThr;
    }else{
      S[i] = 0;
      Thr[i] = (Thr[i]-BaseThr)*decay + BaseThr;
    }
  }
}

void set_MF(double ratio, double rate)
{
  int i, n_mf;
  
  n_mf = N_MF * ratio;

  for(i = 0; i < n_mf; i++){
    if (genrand_real2() < DT*rate){
      S_MF[i] = 1;
    }else{
      S_MF[i] = 0;
    }
  }
  for(i = n_mf; i < N_MF; i++){
    S_MF[i] = 0;
  }
}

void output_spike(FILE *file, double t, int N, int *S)
{
  int i;

  for(i = 0; i < N; i++){
    if (S[i] == 1){
      fprintf(file, "%f %d\n", t, i);
      spike[(int)floor(t+T_Offset)][i] = 1;
    }else{
      spike[(int)floor(t+T_Offset)][i] = 0;
    }
  }
}

void loop(void)
{
  FILE *file_Gr, *file_Go;
  double t = -T_Offset;
  int i;
  int counter = 0;
  
  file_Gr = fopen("gr.dat", "w");
  file_Go = fopen("go.dat", "w");

  while(t < T){

    if (t < 0){
      set_MF(0.05, 0.1);	// 5%, 100Hz
    }else{
      set_MF(0.25, 0.1);	// 5%+20%, 100Hz
    }

    // compute \Delta g for Go<-MF, Go<-Gr, Gr<-MF and GR<-Go
    compute_dg(N_Go, dg_Go_MF,
	       N_Go_MF, id_Go_MF, S_MF, W_Go_MF, g_Go_MF, Tau_Go_MF);
    compute_dg(N_Go, dg_Go_Gr,
	       N_Go_Gr, id_Go_Gr, S_Gr, W_Go_Gr, g_Go_Gr, Tau_Go_Gr);
    compute_dg(N_Gr, dg_Gr_MF,
	       N_Gr_MF, id_Gr_MF, S_MF, W_Gr_MF, g_Gr_MF, Tau_Gr_MF);
    compute_dg(N_Gr, dg_Gr_Go,
	       N_Gr_Go, id_Gr_Go, S_Go, W_Gr_Go, g_Gr_Go, Tau_Gr_Go);

    // g = g + \Delta g
    update_g(N_Go, dg_Go_MF, g_Go_MF);
    update_g(N_Go, dg_Go_Gr, g_Go_Gr);
    update_g(N_Gr, dg_Gr_MF, g_Gr_MF);
    update_g(N_Gr, dg_Gr_Go, g_Gr_Go);

    // compute \Delta V for Go<-MF, Go<-Gr, Gr<-MF and GR<-Go
    compute_dV(N_Go, dV_Go, V_Go, g_Go_leak, g_Go_MF, E_ex, g_Go_Gr, E_ex);
    compute_dV(N_Gr, dV_Gr, V_Gr, g_Gr_leak, g_Gr_MF, E_ex, g_Gr_Go, E_inh);

    // V = V + \Delta V
    update_V(N_Go, dV_Go, V_Go);
    update_V(N_Gr, dV_Gr, V_Gr);

    // error-check block
    for(i = 0; i < N_Go; i++){
      if (isnan(V_Go[i])){
	fprintf(stderr, "err! Go: %d\n", i);
	exit(1);
      }
    }
    for(i = 0; i < N_Gr; i++){
      if (isnan(V_Gr[i])){
	fprintf(stderr, "err! Gr: %d\n", i);
	exit(1);
      }
    }

    // elicit a spike and raise Thr if V > Thr
    generate_spike(N_Go, V_Go, S_Go, Thr_Go, BaseThr_Go, MaxThr_Go, Tau_Go_Thr);
    generate_spike(N_Gr, V_Gr, S_Gr, Thr_Gr, BaseThr_Gr, MaxThr_Gr, Tau_Gr_Thr);

    // output spike trains to a file for each 1 ms
    if (counter == N_Step){
      output_spike(file_Gr, t, N_Gr, S_Gr);
      output_spike(file_Go, t, N_Go, S_Go);
      counter = 1;
    }else{
      counter += 1;
    }

    t += DT;
  }

  fclose(file_Gr);
  fclose(file_Go);
}

int **allocate_connection(int N, int N_pre)
{
  int i, **p;

  p = (int **)malloc(N*sizeof(int *));
  for(i = 0; i < N; i++){
    p[i] = (int *)malloc(N_pre*sizeof(int));
  }
  return p;
}

void initialize_connection(void)
{
  int x, y, n, x_pre, y_pre, n_pre, k, i, j;

  // Unlike the original paper, all connections were set randomly
  // for simplicity.

  for(i = 0; i < N_Go; i++){
    for(n = 0; n < N_Go_Gr; n++){
      k = floor(N_Gr*genrand_real2());
      id_Go_Gr[i][n] = k;
    }
  }
  for(i = 0; i < N_Gr; i++){
    for(n = 0; n < N_Gr_Go; n++){
      k = floor(N_Go*genrand_real2());
      id_Gr_Go[i][n] = k;
    }
  }
  for(i = 0; i < N_Gr; i++){
    for(n = 0; n < N_Gr_MF; n++){
      k = floor(N_MF*genrand_real2());
      id_Gr_MF[i][n] = k;
    }
  }
  for(i = 0; i < N_Go; i++){
    for(n = 0; n < N_Go_MF; n++){
      k = floor(N_MF*genrand_real2());
      id_Go_MF[i][n] = k;
    }
  }

}
void free_connection(int N, int **p)
{
  int i;

  for(i = 0; i < N; i++){
    free(p[i]);
  }
  free(p);
}

void initialize_threshold(void)
{
  int i;

  for(i = 0; i < N_Go; i++){
    Thr_Go[i] = BaseThr_Go;
  }
  for(i = 0; i < N_Gr; i++){
    Thr_Gr[i] = BaseThr_Gr;
  }

}
void initialize_V(void)
{
  int i;

  for(i = 0; i < N_Go; i++){
    V_Go[i] = E_leak;
    dV_Go[i] = 0;
  }
  for(i = 0; i < N_Gr; i++){
    V_Gr[i] = E_leak;
    dV_Gr[i] = 0;
  }
}

int **initialize_spike(int t, int N)
{
  int i, **p;

  p = (int **)malloc(t*sizeof(int *));
  for(i = 0; i < t; i++){
    p[i] = (int *)malloc(N*sizeof(int));
  }
  return p;
}
void free_spike(int t, int **p)
{
  int i;

  for(i = 0; i < t; i++){
    free(p[i]);
  }
  free(p);
}

void initialize(void)
{
  // id_A_B[i][k]: neuron index of the k th connection at postsynaptic neuron i
  // A and B represent post- and pre-synaptic neuron types, respectively.
  id_Go_MF = allocate_connection(N_Go, N_Go_MF);
  id_Go_Gr = allocate_connection(N_Go, N_Go_Gr);
  id_Gr_MF = allocate_connection(N_Gr, N_Gr_MF);
  id_Gr_Go = allocate_connection(N_Gr, N_Gr_Go);

  // spike trains of granule cells for computing PKJ cell activity.
  spike = initialize_spike(T+T_Offset, N_Gr);

  initialize_connection();
  initialize_threshold();
  initialize_V();
}

void finalize(void)
{
  free_connection(N_Go, id_Go_MF);
  free_connection(N_Go, id_Go_Gr);
  free_connection(N_Gr, id_Gr_MF);
  free_connection(N_Gr, id_Gr_Go);
  free_spike(T+T_Offset, spike);
}

void pkj(int US, char *filename)
{
  int t, dt;
  int i;
  double w[N_Gr], pkj, pkj_prev;
  FILE *file;
  const double decay = 0.1;
  const double w_init = 2;

  for(i = 0; i < N_Gr; i++){
    w[i] = w_init;
  }

  file = fopen(filename, "w");

  // learning phase
  pkj_prev = 0;
  for(t = -T_Offset; t < T; t++){
    pkj = 0;
    for(i = 0; i < N_Gr; i++){
      // synaptic weight of a granule cell is set to 0 if the cell
      // elicits at least one spike until 20 ms after the US presentation.
      if (US <= t && t < US+20 && spike[t+(int)T_Offset][i] == 1){
	w[i] = 0;
      }
      // PKJ cell activity is a weighted sum of granule cell activity
      // times syaptic weights
      pkj += w[i]*spike[t+(int)T_Offset][i];
    }
    pkj = decay*pkj + (1-decay)*pkj_prev; // smoothing
    fprintf(file, "%d %f\n", t, pkj);
    pkj_prev = pkj;
  }

  fprintf(file, "\n\n");

  // test phase
  pkj_prev = 0;
  for(t = -T_Offset; t < T; t++){
    pkj = 0;
    for(i = 0; i < N_Gr; i++){
      pkj += w[i]*spike[t+(int)T_Offset][i];
    }
    pkj = decay*pkj + (1-decay)*pkj_prev;
    fprintf(file, "%d %f\n", t, pkj);
    pkj_prev = pkj;
  }

  fclose(file);
}

int main(int argc, char *argv[])
{
  double US;
  if (argc < 2){
    fprintf(stderr, "usage: %s <US>\n", argv[0]);
    exit(1);
  }

  US = atof(argv[1]);
  if (0 <= US && US < T){
    //nop;
  }else{
    fprintf(stderr, "US must be set at [0, %f]\n", T);
    exit(1);
  }
  init_genrand(23L);
  initialize();
  loop();
  pkj(US, "pkj.dat");
  finalize();

  return 0;
}
